import pandas as pd

### DFA to tsv -----------------------------------------------------------------
def read_dfa_contents(dfa_contents):
    nodes = dfa_contents.strip().split('%')

    node_map = {}

    for node in nodes:
        lines = node.strip().split('\n')
        node_id = None
        node_info = {}
        for line in lines:
            if line.startswith("node:"):
                node_id = line.split(":")[1].strip()
            elif line != "":
                key, value = line.split(':')
                value_list = value.strip().split(';')
                if 'R' in key:
                    direction = 'right'
                else:
                    direction = 'left'
                node_info[key.strip()] = {
                    "direction": direction,
                    "next_node": value_list[2],
                    "base": value_list[0],
                    "weight": round(float(value_list[1]), 2),
                }
        if node_id:
            node_map[node_id] = node_info
    
    return node_map

def node_map_to_tsv(aa_seq, utr_trimmed, node_map, file_path):
    node_data = []
    utr_start_pos = (len(aa_seq) - len(utr_trimmed) // 3) * 3

    for node_id, info in node_map.items():
        node_id_first = int(node_id.split(",")[0].strip('('))
        node_id_second = int(node_id.split(",")[1].strip(')'))

        if node_id_first // 3 == len(aa_seq):
            node_name = 'X'
        else:
            node_name = aa_seq[node_id_first // 3]

        for key, details in info.items():
            utr = node_id_first >= utr_start_pos
            user_utr = 'N'

            if node_id_first == utr_start_pos and details["direction"] == 'left':
                utr = False
            if utr:
                utr_index = node_id_first - utr_start_pos
                if utr_index <= len(utr_trimmed):
                    if details["direction"] == 'right':
                        user_utr = utr_trimmed[utr_index]
                    else:
                        user_utr = utr_trimmed[utr_index - 1]

            node_data.append({
                "node_id": node_id,
                "node_aa": f"{node_name}{node_id_first // 3}_{node_id_first % 3}_{node_id_second}",
                "utr": utr,
                "user_utr": user_utr,
                "direction": details["direction"],
                "next_node": details["next_node"],
                "base": details["base"],
                "weight": details["weight"],
            })


    df = pd.DataFrame(node_data)
    df.to_csv(file_path, sep='\t', index=False)

### ----------------------------------------------------------------------------
### graphviz -------------------------------------------------------------------
def generate_graphviz_code(df):
    node_default = {
        "shape": "circle",
        "style": "filled",
        "color": "gainsboro"
    }
    
    graph_code = ["digraph G {\n    rankdir=LR"]
    
    for node_id in df['node_id'].unique():
        node_color = node_default["color"]
        node_aa = df[df['node_id'] == node_id]['node_aa'].values[0]
        node_utr = df[df['node_id'] == node_id]['utr'].values[0]

        if int(node_id.split(',')[0][1:]) % 3 == 0:
            node_color = "aquamarine3"
            if node_utr == True:
                node_color = "cornflowerblue"
        if '*' in node_aa and "_0_" in node_aa:
            node_color = "darksalmon"

        graph_code.append(f'    "{node_id}" [label="{node_id}\\n{node_aa}", shape={node_default["shape"]}, style={node_default["style"]}, color={node_color}];')

    last_node_id = df['node_id'].unique()[-1]
    graph_code.append(f'    "{last_node_id}" [color=darksalmon];\n')
    
    for _, row in df[df['direction'] == 'right'].iterrows():
        edge_label = f'{row["base"]}:{row["weight"]}'

        if row["utr"] and row["user_utr"] != row["base"]:
            graph_code.append(f'    "{row["node_id"]}" -> "{row["next_node"]}" [label="{edge_label}", color=gray];')
        else:
            graph_code.append(f'    "{row["node_id"]}" -> "{row["next_node"]}" [label="{edge_label}"];')

    graph_code.append("}")
    
    return "\n".join(graph_code)

### ----------------------------------------------------------------------------


import argparse
import pandas as pd
import graphviz
from dfa_gen.dfa_class import Codon, DFA, NodeType
from dfa_gen.dfa_util import utr_to_aa, convert_aa_to_triple
from dfa_gen.generate_lattice import prepare_codon_unit_lattice
from dfa_gen.dfa_to_graph import read_dfa_contents, node_map_to_tsv, generate_graphviz_code

def get_dfa(aa_graphs, aa_seq):
    dfa = DFA()
    newnode = NodeType(3 * len(aa_seq), 0)

    dfa.add_node(newnode)

    for i, aa in enumerate(aa_seq):
        i3 = i * 3
        graph = aa_graphs[aa]

        for pos in range(3):
            for node in graph.nodes[pos]:
                num = node.num
                newnode = NodeType(i3 + pos, num)
                
                dfa.add_node(newnode)
                for edge in graph.right_edges[node]:
                    n2 = edge.node
                    nuc = edge.nuc
                    num = n2.num
                    newn2 = NodeType(i3 + pos + 1, num)
                    dfa.add_edge(newnode, newn2, nuc, edge.weight * 100)
    
    return dfa

def dfa_generator(seq, utr="", lambda_val=0):
    SEQ = seq
    UTR = utr
    LAMBDA_VAL = lambda_val
    DFA_FILE = f"test/dfa_{SEQ}"

    CODON_TABLE = "dfa_gen/data/codon_freq_table.tsv"
    CODING_WHEEL = "dfa_gen/data/coding_wheel.txt"

    codon_table = pd.read_csv(CODON_TABLE, sep='\t')
    utr_trimmed = UTR[:len(UTR) - (len(UTR) % 3)]
    utr_aa = utr_to_aa(utr_trimmed, codon_table)
    

    if UTR != "" and SEQ[-1] != '*':
        SEQ = SEQ + '*'
    aa_seq = SEQ + utr_aa

    codon = Codon(CODON_TABLE)
    aa_graphs_with_ln_weights = prepare_codon_unit_lattice(CODING_WHEEL, codon, lambda_=LAMBDA_VAL)
    
    aa_tri_seq = convert_aa_to_triple(aa_seq)
    protein = aa_tri_seq.split()
    dfa = get_dfa(aa_graphs_with_ln_weights, protein)

    with open(DFA_FILE, 'w') as f:
        f.write(f"# {SEQ}\n")
        f.write(f"# {UTR}\n")
        f.write(f"# {aa_seq}\n")
        dfa.print(f)
        
    with open(DFA_FILE, 'r') as file:
        SEQ = file.readline().strip()[2:]
        UTR = file.readline().strip()[2:]
        dfa_input_seq = file.readline().strip()[2:]
        dfa_contents = file.read()

    node_map = read_dfa_contents(dfa_contents)
    node_map_to_tsv(dfa_input_seq, utr_trimmed, node_map, f"{DFA_FILE}.tsv")

    df = pd.read_csv(f"{DFA_FILE}.tsv", sep='\t')
    graphviz_code = generate_graphviz_code(df)
    dot = graphviz.Source(graphviz_code)
    dot.render(f'{DFA_FILE}_graph', format='png')

def main():
    parser = argparse.ArgumentParser(description="Generate DFA from sequence")
    parser.add_argument("seq", type=str, help="The amino acid sequence")
    parser.add_argument("-u", "--utr", type=str, default="", help="The 3'UTR sequence")
    parser.add_argument("-l", "--lambda_val", type=float, default=0, help="The lambda value for calculating edge weight")

    args = parser.parse_args()
    dfa_generator(args.seq, args.utr, args.lambda_val)

if __name__ == "__main__":
    main()


### utils ----------------------------------------------------------------------
def convert_aa_to_triple(aa_seq):
    aa_to_triple = {
        'F': "Phe", 'L': "Leu", 'S': "Ser", 'Y': "Tyr", 
        'C': "Cys", 'W': "Trp", 'P': "Pro", 'H': "His", 
        'Q': "Gln", 'R': "Arg", 'I': "Ile", 'M': "Met",
        'T': "Thr", 'N': "Asn", 'K': "Lys", 'V': "Val", 
        'D': "Asp", 'E': "Glu", 'G': "Gly", 'A': "Ala", 
        '*': "STOP",
    }
    return ' '.join([aa_to_triple[aa] for aa in aa_seq])

def utr_to_aa(utr_seq, codon_table):
    codon_to_aa = dict(zip(codon_table['codon'], codon_table['AA']))
    utr_aa = ""

    for i in range(0, len(utr_seq), 3):
        codon = utr_seq[i:i+3]
        aa = codon_to_aa[codon]
        utr_aa += aa

    return utr_aa

def get_ACGU_num(x):
    return 'A' if x == 1 else 'C' if x == 2 else 'G' if x == 3 else 'U' if x == 4 else 'X'

def get_ACGU_num_nuc(x):
    return 1 if x == 'A' else 2 if x == 'C' else 3 if x == 'G' else 4 if x == 'U' else 0


from setuptools import setup, find_packages

setup(
    name="dfa-gen",
    version="0.1",
    description='Python package for generating mRNA DFA from protein sequence and visualizing it as a graph image',
    packages=find_packages(),
    install_requires=[
        "pandas",
        "graphviz"
    ],
    entry_points={
        "console_scripts": [
            "dfa-gen=dfa_gen.generate_dfa:main",
        ],
    },
)



from collections import defaultdict
from dfa_gen.dfa_class import NodeType, Lattice
from dfa_gen.dfa_util import get_ACGU_num_nuc
import math

### Lattice construction -------------------------------------------------------
def prepare_codon_unit_lattice(wheel_path, codon, lambda_=1.0):
    '''
    Step 1: Read initial weights
    Step 2: Update best weights
    Step 3: Read log weights
    '''
    nodes_with_best_weight = defaultdict(lambda: defaultdict(float))
    edges_with_best_weight = defaultdict(lambda: defaultdict(float))
    
    aa_graphs_with_weights = read_wheel_with_weights(wheel_path, nodes_with_best_weight, edges_with_best_weight, codon)
    update_best_weights(aa_graphs_with_weights, nodes_with_best_weight, edges_with_best_weight)
    aa_graphs_with_ln_weights = read_wheel_with_weights_log(wheel_path, nodes_with_best_weight, edges_with_best_weight, codon, lambda_)

    return aa_graphs_with_ln_weights

def read_wheel_with_weights(filename, nodes_with_best_weight, edges_with_best_weight, codon):
    aa_graphs = {}

    with open(filename) as inFile:
        for line in inFile:
            stuff = line.strip().split('\t')
            aa = stuff[0]
            graph = Lattice()
            graph.add_node(NodeType(0, 0))

            last_first = ''
            i = 0
            for option in stuff[1:]:
                option_splited = option.split(' ')
                first = option_splited[0]
                second = option_splited[1]
                thirds = option_splited[2]
                n2 = NodeType(2, i)
                graph.add_node(n2)
                if first != last_first:
                    n1 = NodeType(1, i)
                    graph.add_node(n1)
                    first_num = get_ACGU_num_nuc(first)
                    weight = 0.0
                    if NodeType(0, 0) in nodes_with_best_weight.get(aa, {}):
                        weight = edges_with_best_weight[aa][(NodeType(0, 0), first_num, n1)] / nodes_with_best_weight[aa][NodeType(0, 0)]
                    graph.add_edge(NodeType(0, 0), n1, first_num, weight)
                else:
                    n1 = NodeType(1, i-1)

                last_first = first
                second_num = get_ACGU_num_nuc(second)
                weight = 0.0

                if n1 in nodes_with_best_weight.get(aa, {}):
                    weight = edges_with_best_weight[aa][(n1, second_num, n2)] / nodes_with_best_weight[aa][n1]
                graph.add_edge(n1, n2, second_num, weight)

                for third in thirds:
                    three_nums = first + second + third
                    weight = 0.0
                    if n2 in nodes_with_best_weight.get(aa, {}):
                        weight = codon.get_weight(aa, three_nums) / nodes_with_best_weight[aa][n2]
                    else:
                        weight = codon.get_weight(aa, three_nums)
                    graph.add_edge(n2, NodeType(0, 0), get_ACGU_num_nuc(third), weight)

                i += 1
            aa_graphs[aa] = graph

    return aa_graphs

def update_best_weights(aa_graphs_with_weights, nodes_with_best_weight, edges_with_best_weight):
    for aa, graph in aa_graphs_with_weights.items():
        for node_at_2 in graph.nodes[2]:
            for node_at_3_nuc_weight in graph.right_edges[node_at_2]:
                node_at_3 = node_at_3_nuc_weight.node
                nuc = node_at_3_nuc_weight.nuc
                weight = node_at_3_nuc_weight.weight
                nodes_with_best_weight[aa][node_at_2] = max(nodes_with_best_weight[aa][node_at_2], weight)
                edges_with_best_weight[aa][(node_at_2, nuc, node_at_3)] = weight

        for node_at_1 in graph.nodes[1]:
            for node_at_2_nuc_weight in graph.right_edges[node_at_1]:
                node_at_2 = node_at_2_nuc_weight.node
                nuc = node_at_2_nuc_weight.nuc
                nodes_with_best_weight[aa][node_at_1] = max(nodes_with_best_weight[aa][node_at_1], nodes_with_best_weight[aa][node_at_2])
                edges_with_best_weight[aa][(node_at_1, nuc, node_at_2)] = nodes_with_best_weight[aa][node_at_2]

        for node_at_0 in graph.nodes[0]:
            for node_at_1_nuc_weight in graph.right_edges[node_at_0]:
                node_at_1 = node_at_1_nuc_weight.node
                nuc = node_at_1_nuc_weight.nuc
                nodes_with_best_weight[aa][node_at_0] = max(nodes_with_best_weight[aa][node_at_0], nodes_with_best_weight[aa][node_at_1])
                edges_with_best_weight[aa][(node_at_0, nuc, node_at_1)] = nodes_with_best_weight[aa][node_at_1]


def read_wheel_with_weights_log(filename, nodes_with_best_weight, edges_with_best_weight, codon, lambda_):
    aa_graphs = {}

    with open(filename) as inFile:
        for line in inFile:
            stuff = line.strip().split('\t')
            aa = stuff[0]
            graph = Lattice()
            graph.add_node(NodeType(0, 0))

            last_first = ''
            i = 0
            for option in stuff[1:]:
                option_splited = option.split(' ')
                first = option_splited[0]
                second = option_splited[1]
                thirds = option_splited[2]
                n2 = NodeType(2, i)
                graph.add_node(n2)
                if first != last_first:
                    n1 = NodeType(1, i)
                    graph.add_node(n1)
                    first_num = get_ACGU_num_nuc(first)
                    weight = 1.0
                    if NodeType(0, 0) in nodes_with_best_weight.get(aa, {}):
                        weight = lambda_ * math.log(edges_with_best_weight[aa][(NodeType(0, 0), first_num, n1)] / nodes_with_best_weight[aa][NodeType(0, 0)])
                    graph.add_edge(NodeType(0, 0), n1, first_num, weight)
                else:
                    n1 = NodeType(1, i-1)

                last_first = first
                second_num = get_ACGU_num_nuc(second)
                weight = 1.0
                if n1 in nodes_with_best_weight.get(aa, {}):
                    weight = lambda_ * math.log(edges_with_best_weight[aa][(n1, second_num, n2)] / nodes_with_best_weight[aa][n1])
                graph.add_edge(n1, n2, second_num, weight)

                for third in thirds:
                    three_nums = first + second + third
                    weight = 1.0
                    if n2 in nodes_with_best_weight.get(aa, {}):
                        weight = lambda_ * math.log(codon.get_weight(aa, three_nums) / nodes_with_best_weight[aa][n2])
                    else:
                        weight = lambda_ * math.log(codon.get_weight(aa, three_nums))
                    graph.add_edge(n2, NodeType(0, 0), get_ACGU_num_nuc(third), weight)

                i += 1
            aa_graphs[aa] = graph

    return aa_graphs

### ----------------------------------------------------------------------------


from collections import defaultdict, namedtuple
from dfa_gen.dfa_util import get_ACGU_num
import csv

NodeType = namedtuple("NodeType", ["index", "num"])
NodeNucWType = namedtuple("NodeNucWType", ["node", "nuc", "weight"])
NodeNucNodeType = namedtuple("NodeNucNodeType", ["node1", "nuc", "node2"])

### ----------------------------------------------------------------------------
k_map_3_1 = {
    "Phe": 'F',
    "Leu": 'L',
    "Ser": 'S',
    "Tyr": 'Y',
    "STOP": '*',
    "Cys": 'C',
    "Trp": 'W',
    "Pro": 'P',
    "His": 'H',
    "Gln": 'Q',
    "Arg": 'R',
    "Ile": 'I',
    "Met": 'M',
    "Thr": 'T',
    "Asn": 'N',
    "Lys": 'K',
    "Val": 'V',
    "Asp": 'D',
    "Glu": 'E',
    "Gly": 'G',
    "Ala": 'A'
}

### (class) Codon --------------------------------------------------------------
class Codon:
    def __init__(self, path):
        self.codon_table = {}
        self.aa_table = defaultdict(list)
        self.max_aa_table = {}

        with open(path) as codon_file:
            reader = csv.DictReader(codon_file, delimiter='\t')
            for row in reader:
                codon = row['codon']
                aa = row['AA']
                fraction = float(row['freq'])
                self.codon_table[codon] = (aa, fraction)
                self.aa_table[aa].append((codon, fraction))
                self.max_aa_table[aa] = max(self.max_aa_table.get(aa, 0), fraction)

            if len(self.codon_table) != 64:
                raise ValueError("Codon frequency file needs to contain 64 codons!")

    def get_weight(self, aa_tri, codon):
        if aa_tri in k_map_3_1:
            aa = k_map_3_1[aa_tri]
            if aa in self.aa_table:
                codons = self.aa_table[aa]
                for c, weight in codons:
                    if c == codon:
                        return weight
        return 0.0

### (class) Lattice ------------------------------------------------------------
class Lattice:
    def __init__(self):
        self.nodes = defaultdict(list)
        self.left_edges = defaultdict(list)
        self.right_edges = defaultdict(list)

    def add_edge(self, n1, n2, nuc, weight=0.0):
        self.right_edges[n1].append(NodeNucWType(n2, nuc, weight))
        self.left_edges[n2].append(NodeNucWType(n1, nuc, weight))

    def add_node(self, n1):
        pos = n1.index
        self.nodes[pos].append(n1)

### (class) DFA ----------------------------------------------------------------
class DFA:
    def __init__(self):
        self.nodes = defaultdict(list)
        self.left_edges = defaultdict(list)
        self.right_edges = defaultdict(list)
        self.auxiliary_left_edges = defaultdict(lambda: defaultdict(list))
        self.auxiliary_right_edges = defaultdict(lambda: defaultdict(list))
        self.node_rightedge_weights = defaultdict(dict)

    def add_edge(self, n1, n2, nuc, weight=0.0):
        self.right_edges[n1].append(NodeNucWType(n2, nuc, weight))
        self.left_edges[n2].append(NodeNucWType(n1, nuc, weight))
        self.auxiliary_right_edges[n1][n2].append((nuc, weight))
        self.auxiliary_left_edges[n2][n1].append((nuc, weight))
        self.node_rightedge_weights[n1][nuc] = weight

    def add_node(self, n1):
        pos = n1.index
        self.nodes[pos].append(n1)

    def print(self, file=None):
        if file is None:
            import sys
            file = sys.stdout
            
        sorted_keys = sorted(self.nodes.keys())
        print("%", file=file)
        for key in sorted_keys:
            for node in self.nodes[key]:
                print(f"node:\t({node.index}, {node.num})", file=file)
                if node in self.right_edges:
                    for cnt, edge in enumerate(self.right_edges[node], 1):
                        print(f"R_{cnt}:\t{get_ACGU_num(edge.nuc)};{edge.weight};({edge.node.index}, {edge.node.num})", file=file)
                if node in self.left_edges:
                    for cnt, edge in enumerate(self.left_edges[node], 1):
                        print(f"L_{cnt}:\t{get_ACGU_num(edge.nuc)};{edge.weight};({edge.node.index}, {edge.node.num})", file=file)
                print("%", file=file)